(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Efuns
open Text

let create_buf_hook = new_local_var_abstr "create_buf_hook"
let modes_alist = new_local_var_abstr "modes_alist"

let create_syntax_table ()  =
  let table = Array.create 256 false 
  in  
  for i = Char.code 'a' to Char.code 'z' do
    table.(i) <- true;
  done;
  for i = Char.code 'A' to Char.code 'Z' do
    table.(i) <- true;
  done;
  for i = Char.code '0' to Char.code '9' do
    table.(i) <- true;
  done;
  table

let default_syntax_table = create_syntax_table ()

let get_name location name =
  let i = ref 0 in
  let compute_name () =
    if !i = 0 then name else
    Printf.sprintf "%s<%d>" name !i
  in
  try
    while true do
      Hashtbl.find location.loc_buffers (compute_name ());
      incr i
    done; assert false
  with
    Not_found -> 
      compute_name ()

let tab_size = ref 9
let create location name text local_map =
  let name = get_name location name in
  let buf =
    { 
    buf_modified = 0;
    buf_text = text;
    buf_name = name;
    buf_filename = None;
    buf_last_saved = version text;
    buf_history = [];
    buf_charreprs = Array.init 256 (fun i ->   String.make 1 (Char.chr i));
    buf_local_map = local_map;
    buf_sync = false;
    buf_mark = None;
    buf_point = Text.add_point text;
    buf_start = Text.add_point text;
    buf_shared = 0;
    buf_syntax_table = default_syntax_table;
    buf_finalizers = [];
    buf_readonly = false;
    buf_local_vars = Vars.empty;
    buf_location = location;
    buf_modes = ["fondamental"];
  } in
  Hashtbl.add location.loc_buffers name buf;
  for i=0 to 25 do
    let s = String.make 2 '^' in
    s.[1] <- Char.chr (97+i);    
    buf.buf_charreprs.(i) <- s
  done;
  buf.buf_charreprs.(9) <- String.make !tab_size ' ';
  exec_hooks (Location location) (fun f -> f buf) create_buf_hook;
  buf

let kill location buf =
  Hashtbl.remove location.loc_buffers buf.buf_name;
  begin
    match buf.buf_filename with
      None -> ()
    | Some filename ->
        Hashtbl.remove location.loc_files filename
  end;
  List.iter (fun f -> f () ) buf.buf_finalizers;
  Gc.compact ();
  buf.buf_shared <- -1

let save buf =
  let filename =
    match buf.buf_filename with
      None -> raise Not_found
    | Some name -> name
  in
  let outc = open_out filename in
  Text.save buf.buf_text outc;
  close_out outc;
  buf.buf_last_saved <- version buf.buf_text
       
exception Found of buffer

let read location filename local_map =
  try
    let filename = 
      if Filename.is_relative filename then
        Filename.concat location.loc_dirname filename
      else
        filename
    in
    try
      Hashtbl.find location.loc_files filename
    with
      Not_found ->
    let name = 
      get_name location (Filename.basename filename) in
    let text =
      try
        let inc = open_in filename in
        let text = Text.read inc in         
        close_in inc; 
        text
      with
        _ -> Text.create ""
    in
    let buf = create location name text local_map in
    Hashtbl.add location.loc_files filename buf;
    buf.buf_filename <- Some filename;
    buf
  with
    Found buf -> buf

let default location name =
  try
    Hashtbl.find location.loc_buffers name
  with
    Not_found ->
      let str = 
        if name = "*help*" then
"Welcome to Efuns, a small demo editor written in Ocaml.

Version is " ^ Version.version ^"
built by "^ Version.builder ^ " " ^ Version.date ^ " 
with
Efuns installation directory : " ^ Version.efunslib ^ "
Ocaml installation directory : " ^ Version.ocamllib ^ "

Fabrice Le Fessant
PARA/SOR Project
INRIA Rocquencourt

Help for Key Bindings: C-h K
See changes in "^ Version.efunslib ^"/Changes
"
        else ""
      in
      create location name (Text.create str) (Keymap.create ())


let compute_representation buf n =
  Text.compute_representation buf.buf_text buf.buf_charreprs n

exception BufferAlreadyOpened

let change_name location buf filename =
  Hashtbl.remove location.loc_buffers buf.buf_name;
  (match buf.buf_filename with
      None -> ()
    | Some filename ->
        Hashtbl.remove location.loc_files filename);
  let filename = 
    if Filename.is_relative filename then
      Filename.concat location.loc_dirname filename
    else
      filename
  in
  try
    Hashtbl.find location.loc_files filename;
    raise BufferAlreadyOpened
  with
    Not_found ->
      let name = get_name location (Filename.basename filename) in
      Hashtbl.add location.loc_buffers name buf;
      Hashtbl.add location.loc_files filename buf;
      buf.buf_filename <- Some filename;
      buf.buf_name <- name
 
  
let set_mark buf point =
  let text = buf.buf_text in
  buf.buf_modified <- buf.buf_modified + 1;
  match buf.buf_mark with
    None ->
      let mark = dup_point text point in
      buf.buf_mark <- Some mark
  | Some mark ->
      goto_point text mark point

let rec get_mark buf point =
  match buf.buf_mark with
    None -> 
      set_mark buf point;
      get_mark buf point
  | Some mark -> mark

let remove_mark buf =
  match buf.buf_mark with
    None -> ()
  | Some mark ->
      buf.buf_mark <- None;
      remove_point buf.buf_text mark;
      buf.buf_modified <- buf.buf_modified + 1

let modes_old = ref []
let regexp_alist = ref []

let set_buffer_mode buf =
  let buf_name = 
    match buf.buf_filename with
      None -> buf.buf_name
    | Some file_name -> file_name 
  in
  let modes_alist = get_local_var (Buffer buf) modes_alist in
  if not (!modes_old == modes_alist) then
    begin
      regexp_alist := 
         List.map 
           (fun (file_reg, install_major) ->
             Str2.regexp file_reg, install_major) modes_alist;
      modes_old := modes_alist;
    end;
  try
    List.iter (fun (regexp, install) ->
      if Str2.string_match regexp buf_name 0 then
          try
            (install buf : unit);
            raise Exit
          with
            _ -> raise Exit
      ) !regexp_alist
  with
    Exit -> ()

let set_major_mode buf name =
  buf.buf_modified <- buf.buf_modified + 1;
  match buf.buf_modes with
    major :: modes -> buf.buf_modes <- name :: modes
  | [] -> buf.buf_modes <- [name]

let set_minor_mode buf name =
  buf.buf_modes <- buf.buf_modes @ [name];
  buf.buf_modified <- buf.buf_modified + 1

let del_minor_mode buf name =
  buf.buf_modified <- buf.buf_modified + 1;
  buf.buf_modes <- List.rev 
       (List.fold_left 
          (fun tail mode -> 
            if mode = name then tail else (mode :: tail)) [] buf.buf_modes)

let modep buf name =
  List.mem name buf.buf_modes

let _ =
  Efuns.add_start_hook 
    (fun location ->
      set_local_var (Location location) create_buf_hook [set_buffer_mode];
      set_local_var (Location location) modes_alist []
      )
