(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)


open Efuns
open Text
open Frame
open Simple
open Select
open Multi_frames  

let compilation_frame = ref None

type error = {
(* error location *)    
    err_filename : string;
    err_line : int;
    err_begin : int;
    err_end : int;
(* error message *)
    err_msg : int;
  }

let c_error_regexp = Str2.regexp "^\([^:]+\):\([0-9]+\):.*$"

let c_find_error text error_point =
  let groups = 
    Text.search_forward_groups text c_error_regexp 
      error_point 2 in
  let error =
    {  
      err_msg = Text.get_position text error_point;
      err_filename = groups.(0);
      err_line = (int_of_string groups.(1)) - 1;
      err_begin = 0;
      err_end = 0;
    } in
  Text.fmove text error_point 1;
  error


let find_error = new_local_var_abstr "find_error"

let next_error top_frame =
  let top_window = Window.top top_frame.frm_window in
  let location = top_window.top_location in
  match !compilation_frame with
    None -> Top_window.message top_window "No compilation started"
  | Some (frame, error_point, cdir) ->      
      if frame.frm_killed then unkill (cut_frame top_frame) frame;
      let buf = frame.frm_buffer in
      let find_error =
        try
          get_local_var (Buffer buf) find_error
        with
          Failure _ -> c_find_error
      in
      let text = buf.buf_text in
      let point = frame.frm_point in
      try
        let error = find_error text error_point in
        Text.set_position text frame.frm_start error.err_msg;
        Text.set_position text point error.err_msg;
        frame.frm_force_start <- true;
        frame.frm_redraw <- true;
        if error.err_filename <> "" then
          let filename = Filename.concat cdir error.err_filename in
          let buf = Buffer.read location filename (Keymap.create ()) in
(* new frame for error buffer *)
          let frame = 
            try
              find_buffer_frame location buf
            with
              Not_found ->
                if frame == top_frame then
                  let display = Window.display top_window in
                  let new_window = 
                    Top_window.create top_window.top_location 
                      display 
                  in
                  Frame.create new_window.top_windows None buf
                else
                  Frame.create top_frame.frm_window None buf
          in
          let text = buf.buf_text in
          let point = frame.frm_point in
          Text.point_to_line text point error.err_line;
          Text.fmove text point error.err_begin;
          Frame.active frame
      with
        Not_found ->
          Top_window.message top_window "No more errors"


let make_hist = ref ["make -k"]
let compile find_error_fun frame =
  let default = List.hd !make_hist in
  select_string frame ("Compile command: (default :"^ default^") " )
  make_hist ""
    (fun cmd -> 
      let cmd = 
        if cmd = "" then default else
          cmd 
      in
      let top_window = Window.top frame.frm_window in
      let location = top_window.top_location in
      let comp_window =
        match !compilation_frame with
          None -> cut_frame frame 
        | Some (new_frame,error_point, _) ->
            Text.remove_point new_frame.frm_buffer.buf_text error_point;
            Buffer.kill location new_frame.frm_buffer;
            if new_frame.frm_killed then cut_frame frame
            else new_frame.frm_window 
      in
      let cdir = Frame.current_dir frame in
      Unix.chdir cdir;
      let comp_frame = System.start_command comp_window cmd in
      Frame.active frame;
      let buf = comp_frame.frm_buffer in
      let error_point = Text.add_point buf.buf_text in
      compilation_frame := Some (comp_frame, error_point, cdir);
      set_local_var (Buffer buf) find_error find_error_fun
  )

let set_compilation_buffer frame comp_buf cdir =
  let error_point = add_point comp_buf.buf_text in
  let window =
    match !compilation_frame with
      None -> 
        cut_frame frame
    | Some (frame,point, _) ->
        remove_point frame.frm_buffer.buf_text point;  
        if frame.frm_killed then cut_frame frame 
        else
          frame.frm_window
  in
  let error_point = add_point comp_buf.buf_text in
  let comp_frame = Frame.create window None comp_buf in
  compilation_frame := Some (comp_frame, error_point, cdir)
  
  