(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* Useful types for Efuns *)

open Misc
open Obj
  
exception UnboundKey

module Vars = Map.Make (struct  
  type t = string
  let compare = compare end)

type local_vars = Obj.t ref Vars.t

type map =
    { char_map : binding array;
      mutable complex_bindings : (key * binding) list;
      mutable partial_map : bool;
    } 

and key = mod_ident * Xtypes.keySym

and action = frame -> Obj.t

and mod_ident =
    NormalMap
  | ControlMap
  | MetaMap

and prefix = key list

and binding = 
    Function of action
  | Prefix of map
  | Unbound

(* a buffer containing a file in Text.t *)
and buffer =
    { 
    mutable buf_text : Text.t;
    mutable buf_modified : int;

    mutable buf_name : string;
    mutable buf_filename : string option;
    mutable buf_last_saved : int;
    
    mutable buf_history : (int * Text.action) list;
    
    mutable buf_charreprs : string array;
    mutable buf_syntax_table : bool array;
      
    buf_local_map : map;
    mutable buf_sync : bool;
    mutable buf_mark : Text.point option;
    mutable buf_point : Text.point;
    mutable buf_start : Text.point;
    mutable buf_shared : int; (* number of frames for that buffer *)
    mutable buf_readonly : bool;
    mutable buf_modes : string list;
    mutable buf_finalizers : (unit -> unit) list;

    mutable buf_local_vars : local_vars;
    buf_location : location;
  } 

(* a frame: a view of a buffer for a window *)
and frame  =
    {
    mutable frm_buffer : buffer;
    mutable frm_location : location;
    mutable frm_window : window;
    mutable frm_last_text_updated : int;
    mutable frm_last_buf_updated : int;

    mutable frm_map : map;
    mutable frm_prefix : key list;

    mutable frm_repeat_action : int;
    mutable frm_last_action : action;

    mutable frm_start : Text.point;(* first point of the first buffer-line on screen *)
    mutable frm_end : Text.point; (* last point on screen, -1 if modified *)
    mutable frm_y_offset : int; (* offset(+/-) of screen-lines after frm_start *)
    mutable frm_point : Text.point; (* insert point *)

    mutable frm_cursor_x : int;
    mutable frm_cursor_y : int;
    mutable frm_cursor : string;
    mutable frm_cursor_attr : Text.attribute;

    mutable frm_force_point : bool;
    mutable frm_force_start : bool;
    mutable frm_force_cursor : bool;

    mutable frm_x_offset : int;
    mutable frm_cutline : int; (* max_int for no, else length *)

    mutable frm_has_scrollbar : int;(* 0 for no scrollbar, 2 for scrollbar *)
    mutable frm_has_status_line : int;(* 0 for minibuffer, 1 for normal frame *)
    mutable frm_status : status;    
    
    mutable frm_xpos : int;
    mutable frm_ypos : int;
    mutable frm_width : int;
    mutable frm_height : int;
    
    mutable frm_table : line_repr array;
    mutable frm_killed : bool;
    mutable frm_mini_buffer : string option;
    mutable frm_redraw : bool;

    mutable frm_scrollbar : Xterm.xterm_scrollbar option;
    mutable frm_local_vars : local_vars;
  } 

and status_info =
    StatModified
  | StatName
  | StatLine
  | StatCol
  | StatFile
  | StatMode

and status =
    { mutable status_string : string;
      mutable status_modified : bool;
      mutable status_format : (status_info * (int * int)) list;
      mutable stat_col : int;
      mutable stat_name : string;
      mutable stat_file : string;
      mutable stat_line : int;
      mutable stat_modified : bool;
      mutable stat_modes : string list;
    }

and line_repr =
    { 
    mutable repr_line : Text.line;
    mutable repr_y : int;
    mutable repr_prev_offset : int;
    mutable repr_prev_reprs : Text.repr list;
    mutable repr_offset : int;
    mutable repr_reprs : Text.repr list;
  } 

(* an xterm: a window containing some frames *)
and top_window = 
    { mutable top_location : location;
      mutable top_display : Xterm.xterm_display option;
      mutable top_xterm : Xterm.xterm_window option;
      mutable top_windows : window;
      mutable top_mini_buffers : frame list;
      mutable top_width : int;
      mutable top_height : int;
      mutable top_name : string;
      mutable top_active_frame : frame;
      mutable top_second_cursor : frame option;
    } 

and window =
    { 
    mutable win_xpos : int;
    mutable win_ypos : int;
    mutable win_width : int;
    mutable win_height : int;
    mutable win_down : window_down;
    mutable win_up : window_up;
    mutable win_mini : bool;
    } 

and window_up =
    Window of window
  | TopWindow of top_window

and window_down =
  | HComb of window * window 
  | VComb of window * window
  | NoFrame of unit
  | WFrame of frame

and location =
    { 
    loc_global_map : map;
    mutable loc_windows : top_window list;
    mutable loc_buffers : (string, buffer) Hashtbl.t;
    mutable loc_files : (string, buffer) Hashtbl.t;
    mutable loc_dirname : string;
    mutable loc_width : int;
    mutable loc_height : int;
    mutable loc_fg : string;
    mutable loc_bg : string;
    mutable loc_font : string;
    mutable loc_local_vars : local_vars;
    mutable loc_counter : int;
  
    loc_vars_table : (string, (Obj.t -> string) * (string -> Obj.t)) Hashtbl.t;

    loc_fonts : (string,int) Hashtbl.t;
    loc_fonts_names : string array;
    mutable loc_fonts_n : int;
    loc_colors : (string,int) Hashtbl.t;
    loc_colors_names : string array;
    mutable loc_colors_n : int;
    
    loc_mutex : Concur.Mutex.t;
    } 

type sens = 
    Backward | 
    Forward
type to_regexp =
    Regexp
  | RegexpString


(* Les variables importantes dans le reste du programme. *)

let efunspath =
  try
    let path = Sys.getenv "EFUNSPATH" in
    Misc.string_to_path path
  with
    Not_found -> []

let camldir =
  try
    Sys.getenv "CAMLDIR" 
  with
    Not_found -> Version.ocamllib

let path = (homedir :: (Filename.concat homedir ".efuns") 
  :: (Sys.getcwd ()) :: efunspath) @ 
  [camldir ; Version.efunslib ; Version.ocamllib]

let load_path = ref path

let init_files = ref []
let displayname = ref ""
let width = ref 80
let height = ref 27
let font = ref "fixed"
let fg = ref "white"
let bg = ref "black"

  
let ns_name = ref "newclient"
let ns_host = ref (Unix.gethostname ())
let ns_port = ref 20001

let _ =
  Arg.parse [
  "-display", Arg.String(fun s -> displayname :=s),"<dpy>: Name of display";
  "-fg", Arg.String(fun s -> fg :=s), "<color>: Foreground color";
  "-bg", Arg.String(fun s -> bg :=s), "<color>: Background color";
  "-font", Arg.String(fun s -> font :=s), "<font>: Font name";
  "-loadpath",Arg.String (fun s -> load_path := 
        (Misc.string_to_path s) @ !load_path), "<path>: Load Path";
      "-ns_host", Arg.String (fun s -> ns_host := s), ("<host>: host hosting the name server default: "^ !ns_host);
  "-ns_port", Arg.Int (fun i -> ns_port := i), ("<port>: connect to <port> default: "^ (string_of_int !ns_port));
  "-ns_name", Arg.String (fun s -> ns_name := s), ("<name>: query <name> default: "^ !ns_name);

] (fun name -> init_files := name :: !init_files) 
  "A small editor entirely written in Objective Caml 
by Fabrice LE FESSANT, INRIA Rocquencourt, FRANCE
  http://pauillac.inria.fr/~lefessan/src

Options:
"

(* Les hooks de lancement apres le chargement d'un module *)
  
let start_hooks = ref []
let add_start_hook hook = start_hooks := hook :: !start_hooks

let init (location: location) =
  let rec iter hooks =
    match hooks with
      [] -> ()
    | (f: location -> unit) :: hooks -> 
        f location;
        iter hooks
  in
  let hooks = List.rev !start_hooks in
  start_hooks := [];
  iter hooks

  (* Les variables locales *)

type 'a local_var = string
type env = 
    Buffer of buffer
  | Frame of frame
  | Location of location

let vars_table = Hashtbl.create 203
  
let new_local_var (name: string) string_to_var var_to_string = 
  try
    Hashtbl.find vars_table name; 
    failwith (Printf.sprintf "A variable named %s already exists" name)
  with
    Not_found ->
      Hashtbl.add vars_table name (magic var_to_string,magic string_to_var);
      name

let no_get a = "<abstr>"
let no_set (s:string) = failwith "This variable can not be set"
let new_local_var_abstr name = new_local_var name no_get no_set
  
let set_local_var env var value =
  let value = repr value in
  let set = 
    match env with
      Location location -> location.loc_local_vars
    | Buffer buf -> buf.buf_local_vars
    | Frame frm -> frm.frm_local_vars 
  in
  try
    let r = Vars.find var set in
    r := value
  with
    Not_found ->
      match env with
        Location location -> 
          location.loc_local_vars <- Vars.add var (ref value) set
      | Buffer buf -> buf.buf_local_vars <- Vars.add var (ref value) set
      | Frame frm -> frm.frm_local_vars  <- Vars.add var (ref value) set
             
let rec get_local_var env var =
  match env with
    Frame frm -> 
      (try
        Obj.magic !(Vars.find var frm.frm_local_vars)
      with
        Not_found -> 
          get_local_var (Buffer frm.frm_buffer) var)
  | Buffer buf ->
      (try
        Obj.magic !(Vars.find var buf.buf_local_vars)
      with
        Not_found -> 
          get_local_var (Location buf.buf_location) var)
  | Location location ->
      try
        Obj.magic !(Vars.find var location.loc_local_vars)
      with
        Not_found ->
          failwith (var ^ " not found in local vars")

          (* comment implementer le set-variable d'Emacs ? Ici, les variables ne sont pas typees (du moins, dans la table de stockage). *)
  
          
let add_hook env hook f =
  let hooks =
    try
      get_local_var env hook
    with
      Failure _ -> []
  in
  set_local_var env hook (f::hooks)

let get_hooks env hook =
  try
    get_local_var env hook
  with
    Failure _ -> []

let exec_hooks env compose hook =
  let rec iter hooks =
    match hooks with
      [] -> ()
    | f :: hooks -> 
        compose f;
        iter hooks
  in
  iter (List.rev (get_hooks env hook))
