(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Efuns
open Window

let message top_window msg =
  let xterm = Window.xterm top_window  in
  let len = String.length msg in
  Xterm.draw_string xterm 0 (top_window.top_height - 1)
    msg 0 len Text.direct_attr;
  Xterm.clear_eol xterm len (top_window.top_height - 1)
    (top_window.top_width - len);
  match top_window.top_mini_buffers with
    [] -> (* No mini-buffer ... *) ()
  | mini_buffer :: _ -> (* one mini-buffer is active *)
      match top_window.top_display with
        None -> ()
      | Some display ->
          Xterm.update_displays ();
          Unix.select [] [] [] 0.2;
          mini_buffer.frm_redraw <- true

let clear_message top_window =
  match top_window.top_mini_buffers with
    [] -> 
      let xterm = Window.xterm top_window in
      Xterm.clear_eol xterm 0 
        (top_window.top_height - 1) top_window.top_width; 
  | _ -> ()



let dummy_action frame = ()

let try_map frame map key =
  match Keymap.get_binding map key with
    Unbound -> raise UnboundKey
  | Prefix map ->
      frame.frm_map <- map;
      frame.frm_prefix <- key :: frame.frm_prefix;
      let top_window = top frame.frm_window in
      message top_window (Keymap.print_key_list frame.frm_prefix);
  | Function f ->
      frame.frm_repeat_action <- 
      if (f == frame.frm_last_action) then
        frame.frm_repeat_action + 1
      else 0;
      f frame;
      frame.frm_last_action <- f;
      frame.frm_map <- frame.frm_buffer.buf_local_map;
      frame.frm_prefix <- []

let set_cursor_on top_window frame = 
  Frame.set_cursor frame;
  if frame.frm_cursor.[0] <> '\000' then
    let xterm = xterm top_window in
    Xterm.draw_string xterm 
      (frame.frm_xpos + frame.frm_cursor_x) 
    (frame.frm_ypos + frame.frm_cursor_y) 
    frame.frm_cursor 0 1 Text.inverse_attr

let set_cursor_off top_window frame =
  if frame.frm_cursor.[0] <> '\000' then
    let xterm = xterm top_window in
    Xterm.draw_string xterm 
      (frame.frm_xpos + frame.frm_cursor_x) 
    (frame.frm_ypos + frame.frm_cursor_y) 
    frame.frm_cursor 0 1 frame.frm_cursor_attr

let cursor_on top_window =
  let frame = top_window.top_active_frame in
  let name = frame.frm_buffer.buf_name in
  if not (name == top_window.top_name) then
    begin
      let xterm = xterm top_window in
      Xterm.set_name xterm name;
      top_window.top_name <- name
    end;
  set_cursor_on top_window frame;
  match top_window.top_second_cursor with
    None -> ()
  | Some frame -> set_cursor_on top_window frame

let cursor_off top_window =
  let frame = top_window.top_active_frame in
  set_cursor_off top_window frame;
  match top_window.top_second_cursor with
    None -> ()
  | Some frame -> set_cursor_off top_window frame


let update_display location =
  List.iter (fun top_window -> 
      iter (Frame.update top_window) top_window.top_windows;
      (match top_window.top_mini_buffers with
          [] -> ()
        | frame :: _ ->
            Frame.update top_window frame);
      cursor_on top_window;
  ) location.loc_windows

let clean_display location =
  List.iter (fun top_window -> 
      cursor_off top_window
  ) location.loc_windows

let rec resize_window window xpos ypos width height =
  let old_width = window.win_width in
  let old_height = window.win_height in
  window.win_xpos <- xpos;
  window.win_ypos <- ypos;
  window.win_width <- width;
  window.win_height <- height;
  match window.win_down with
    WFrame frame -> Frame.install window frame
  | NoFrame () -> assert false
  | HComb (w1,w2) ->
      let wi1 = w1.win_width * width / old_width in
      resize_window w1 xpos ypos wi1 height;
      resize_window w2 (xpos + wi1) ypos (width - wi1) height
  | VComb (w1,w2) ->
      let he1 = w1.win_height * height / old_height in
      resize_window w1 xpos ypos width he1;
      resize_window w2 xpos (ypos + he1) width (height - he1)

let rec find_frame window x y =
  match window.win_down with
    NoFrame () -> assert false
  | WFrame frame -> frame
  | HComb (w1,w2) -> 
      if w2.win_xpos > x then find_frame w1 x y
      else find_frame w2 x y
  | VComb (w1,w2) ->
      if w2.win_ypos > y then find_frame w1 x y
      else find_frame w2 x y

let keypressed = ref 0
let mouse_x = ref 0
let mouse_y = ref 0

let find_selected_frame top_window =
  let x = !mouse_x in
  let y = !mouse_y in
  let frame = 
    if y > top_window.top_height - 2 then
      match top_window.top_mini_buffers with
        [] -> raise Not_found
      | mini_frame :: _ -> mini_frame
    else
      find_frame top_window.top_windows x y
  in
  Frame.active frame;
  frame


let mouse_set_active top_window =
  let x = !mouse_x in
  let y = !mouse_y in
  let frame = find_selected_frame top_window in
  Frame.move_point frame frame.frm_point x y;
  frame


let mini_message frame msg =
  let top_window = Window.top frame.frm_window in
  message top_window msg

let handle_key_start_hook = new_local_var_abstr "handle_key_start_hook"
let handle_key_end_hook = new_local_var_abstr "handle_key_end_hook"
  
let handle_key top_window modifiers keysym =
  keypressed := keysym;
  let location = top_window.top_location in
  let frame = top_window.top_active_frame in
  clean_display location;
  clear_message top_window;
  exec_hooks (Location location) (fun f -> f location) handle_key_start_hook;
  begin
    let mods = 
      if modifiers land Xconst.controlMask <> 0 then ControlMap
      else
        if modifiers land Xconst.mod1Mask <> 0 then MetaMap
        else NormalMap in
    let key = (mods, keysym) in
    try
      try
        try_map frame frame.frm_map key
      with
        UnboundKey -> 
          if frame.frm_map.partial_map then
            try_map frame top_window.top_location.loc_global_map
              key
    with
      UnboundKey -> 
        message top_window
          (Printf.sprintf "Unbound key %s %s"
            (Keymap.print_key_list frame.frm_prefix)
          (Keymap.print_key key));
        frame.frm_prefix <- [];
        frame.frm_map <- frame.frm_buffer.buf_local_map
    | Failure str ->
        message top_window str
    | e -> 
        message top_window 
          (Printf.sprintf "Uncaught exception %s"
            (Printexc.to_string e));
  end;
  exec_hooks (Location location) (fun f -> f location) handle_key_end_hook;
  update_display top_window.top_location

let handler top_window xterm event =
  let location = top_window.top_location in
  Concur.Mutex.lock location.loc_mutex;
  try
    begin
      match event with
        Xterm.XTKeyPress (modifiers, s, keysym) ->
          if (keysym < XK.xk_Shift_L || keysym > XK.xk_Hyper_R)
          then 
            handle_key top_window modifiers keysym
      
      | Xterm.XTButtonPress (modifiers,button,x,y) -> 
          mouse_x := x;
          mouse_y := y;
          handle_key top_window modifiers (XK.xk_Pointer_Button_Dflt + button)
      
      | Xterm.XTMouseMotion (modifiers,button,x,y) ->
          mouse_x := x;
          mouse_y := y;
          handle_key top_window modifiers (XK.xk_Pointer_Drag_Dflt + button)
      
      | Xterm.XTResize (new_width, new_height) ->
          resize_window top_window.top_windows 0 0 new_width (new_height - 1);
          List.iter
            (fun frame -> 
              let window = frame.frm_window in
              window.win_ypos <- new_height - 1;
              window.win_width <- new_width;
              Frame.install window frame) top_window.top_mini_buffers;
          top_window.top_width <- new_width;
          top_window.top_height <- new_height;
          clear_message top_window;
          update_display top_window.top_location
          
    end;
    Concur.Mutex.unlock top_window.top_location.loc_mutex;
  with
    e ->   
      Concur.Mutex.unlock top_window.top_location.loc_mutex;
      raise e


let create location display =
  let buf = Buffer.default location "*help*" in
  let xterm = Xterm.create_window 
      display "new_frame"
      location.loc_font location.loc_width
      location.loc_height in
  let window = Window.create_at_top  0 0 location.loc_width (location.loc_height - 1) in
  let frame = Frame.create_without_top location window None buf  in
  let top_window =
    { top_location = location;
      top_display = Some display;
      top_xterm = Some xterm;
      top_windows = window;
      top_mini_buffers = [];
      top_width = location.loc_width;
      top_height = location.loc_height;
      top_name = "window";
      top_active_frame = frame;
      top_second_cursor = None;
    } 
  in
  frame.frm_window.win_up <- TopWindow top_window;
  Xterm.install_handler display xterm (handler top_window xterm);
  location.loc_windows <- top_window :: location.loc_windows;
  top_window


let delete_window frame =
  let top_window = Window.top frame.frm_window in
  let location = top_window.top_location in
  if List.length location.loc_windows > 1 then
    let xterm = Window.xterm top_window in
    Xterm.destroy_window xterm;
    Frame.kill_all top_window.top_windows;
    location.loc_windows <- Misc.list_remove location.loc_windows
      top_window
    
    
