(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Concur
open ThreadUnix
open Xbuffer
open Xtypes
open Conv_event
open Ll_trans
open Xconst

let xopcode opcode = xopcodes.(Obj.magic (opcode : requestOpcode))
let xerror error = xerrors.(Obj.magic (error : errorCode))

let print_xerror error =
    Printf.printf "Warning: Uncaught X Error: %s for request %s\n" (xerror error.err_code) (xopcode error.err_request_code);
  Printf.printf "    Resource: %d  Minor code: %d" error.err_resourceid error.err_minor_code;
  print_newline ()

(**************************************************************

           Le thread de lecture 

***************************************************************)

let readWholeReply socket =
  let b = String.create (4*8)
  in
    sock_io read socket 0 b 8;
    Xdebug.print ("reply type: "^(string_of_int (Char.code b.[0])));
    if b.[0]='\001' then
    let len = getCard32 b 4 
    in
    Xdebug.print ("reply length: "^(string_of_int len));
    if len=0 then
      b (* la reponse attendue *)
    else (* il faut lire le reste *)
      let b2 = newBuffer (len+8)
      in
      string_blit b 0 b2 0 32;
      sock_io read socket 8 b2 len;
      b2
    else
      b
  
let readReply display =
  let b = readWholeReply display.socket
  in
    match b.[0] with
	'\000' (* une erreur *)
      | '\001' -> (* une reponse *)
	  let request = getCard16 b 2 
	  in
	  let rec iter () =
	      let (serial,wrapper) = Xfifo.read display.wrappers_queue
	      in
		if serial > request then
		  raise Xfifo.Empty
		else
                  begin
                    Xfifo.take display.wrappers_queue;
		    if serial = request then
		      wrapper
		    else
		      begin
                        (try
                          wrapper " " (* pas de reponse *)
                        with
                          _ -> (* discard exception *)());
		        iter ()
		      end
                  end
	  in
          Mutex.lock display.dpy_mutex;
	  (try
	    iter () b
	  with
	    Xfifo.Empty when b.[0] = '\000' -> (* Warning for error *)
              let error = getXError b in
              print_xerror error
          | _ -> (* discard exception *) ());
          Mutex.unlock display.dpy_mutex
      | _ -> (* un event *)
	  display.serial_in <- display.serial_in + 1;
          Mutex.lock display.dpy_mutex;
	  Xfifo.put display.event_queue
	    (convertCore2Event display (b,display.serial_in));
          Condition.signal display.dpy_wait; (* signal for nextEvent *)
          Mutex.unlock display.dpy_mutex
  
	  
(*************************************************************

     Les fonctions pour les autres threads

**************************************************************)

let send_alone display buf =
  Mutex.lock display.dpy_mutex;
  display.serial_out <- (display.serial_out + 1) mod 65536;
  let serial =  display.serial_out
  in
    sock_io write display.socket 0 buf (((String.length buf)+1)/4);
    Mutex.unlock display.dpy_mutex;
    serial

let send_with_wrapper display buf wrapper =
  Mutex.lock display.dpy_mutex;
  display.serial_out <- (display.serial_out + 1) mod 65536;
  let serial =  display.serial_out
  in
    Xfifo.put display.wrappers_queue (serial,wrapper);
    sock_io write display.socket 0 buf (((String.length buf)+1)/4);
    Mutex.unlock display.dpy_mutex;
    ()


(************************************************************

        La connection avec le serveur X

*************************************************************)

let closeDisplay display =
  Mutex.lock display.dpy_mutex;
  Concur.Thread.remove_reader display.socket;  (* kill display reader *)
  Unix.close display.socket


let openDisplay name =
  let name =
    if name = "" then 
      try
        Sys.getenv "DISPLAY" 
      with
        Not_found -> ":0"
    else name 
  in
  let (sock,server_name,screen_num,auth) = openConnectionWithDisplay name
  in
  let (xauth_name,xauth_data) = auth in
  let len = 3 +(strLen (String.length xauth_name))+
            (strLen (String.length xauth_data)) in
  let b = newBuffer len  in
    setCard8 b 0 0x6C; (* pour PC *)
    setCard16 b 2 11; (* protocol major *)
    setCard16 b 4 0;  (* protocol minor *)
    setCard16 b 6 (String.length xauth_name);  (* auth protocol *)
    setCard16 b 8 (String.length xauth_data);  (* auth data *)
   (* 10:16 unused *)
    if xauth_name<>"" then setString b 12 xauth_name;
    if xauth_data<>"" then 
      setString b (12+4*(strLen (String.length xauth_name))) xauth_data;
    sock_io write sock 0 b len;
    sock_io read sock 0 b 2;
    let len = getCard16 b 6 in
    let add = newBuffer (len+1) in
    sock_io read sock 0 add len;
    if b.[0]='\000' then
      failwith 
        (Printf.sprintf"Connection refused by %s : %s" server_name (getString add 0 (getCard8 b 1)))
    else
      if b.[0]='\002' then
          failwith ("authentification required:"^add)
      else
        if b.[0] = '\001' then
          begin
            let (maj,min)=(getCard16 b 2,getCard16 b 4)
            in
            let b = add
	    in
            
              let display  =
	      	getXServerInfo b sock maj min screen_num server_name
	      in
	      Thread.add_reader display.socket
                (function () -> readReply display);
	      display
          end
        else
          failwith "Reponse incoherente"

let alloc_id display =
  let incr = ref (display.resource_incr)
  in
  let id = ref (display.last_resource_id+display.resource_incr)
  in 
    while (!id land display.resource_mask) <= (display.last_resource_id) do
        incr := (!id lor display.resource_mask) lxor display.resource_mask;
      id := (!id)+(!incr)
    done;
    id := (!id lor (!incr - 1)) lxor (!incr - 1);
    display.last_resource_id <- !id;
    !id lor display.resource_base

  

let newRequest (op : requestOpcode) len =
  let b = newBuffer len
  in
  setEnum8 b 0 op;
  setCard16 b 2 len;
  b

let newWinRequest (op : requestOpcode) len win =
  let b = newBuffer len
  in
  setEnum8 b 0 op;
  setCard16 b 2 len;
  setCard32 b 4 win;
    b

let emptyRequest mask =
  newRequest mask 1

let simpleRequest mask data =
  let b = newRequest mask 2
  in
    setCard32 b 4 data;
    b

let doubleRequest mask data1 data2 =
  let b = newRequest mask 3
  in
    setCard32 b 4 data1;
    setCard32 b 8 data2;
    b

let littleRequest opcode data =
  let b = emptyRequest opcode
  in
  setEnum8 b 1 data;
  b
               
let parse_buffer event wrapper buffer =
  try
    if buffer.[0]='\000' then
      Jeton.signal_exception event (XError (
                                      let b = getXError buffer in
                                        if !Xdebug.debug_flag then
                                          print_xerror b;
                                        b))
    else
      Jeton.signal_value event (wrapper buffer)
  with
      e -> Jeton.signal_exception event e
