{
(***********************************************************************)
(*                                                                     *)
(*                             Xlib                                    *)
(*                                                                     *)
(*       Fabrice Le Fessant, projets Para/SOR, INRIA Rocquencourt      *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*
  Resources management. Un peu trop complique pour l'instant.
  
  *)

  open Printf
  
  type expr = 
    STAR
  | SEMI
  | POINT
  | EOF
  | NAME of string
  | ERROR
  
  }

rule first = parse
    "!" _ * "\n" { first lexbuf }
  | " " | "\011" { first lexbuf }
  | '\n' { first lexbuf }
  | "#include" _ * '\n'
    { (* not implemented yet  *)
      first lexbuf }
  | '*'  { STAR }
  | [ ^ '.' '*' ' ' '\n' ':'] + { NAME (Lexing.lexeme lexbuf)}
  | eof { EOF }
  | _  { first lexbuf }
    
and next = parse
    ':' { SEMI }
  | '.' { POINT }
  | '*' { STAR }
  | [ ^ '.' '*' ' ' '\n' ':' ] + { NAME (Lexing.lexeme lexbuf)}
  | _ { ERROR }

and ressource = parse
  ([ ^ '\n'] | "\\\n" ) * { Lexing.lexeme lexbuf }

{

type tree =
  Leaf of string
| Branch of (int * tree) list ref
  
type t =
  {
    mutable tags : int; (* 0 = *, 1 = end *)
    names : (string, int) Hashtbl.t;
    table : (int, tree) Hashtbl.t;
  }
  
let create () = 
  let t =
    { tags = 2;
      names = Hashtbl.create 19;
      table = Hashtbl.create 19
    }
  in
  Hashtbl.add t.names "*" 0;
  t
  
let add_name t name =
  try
    Hashtbl.find t.names name
  with
    Not_found ->
      t.tags <- t.tags + 1;
      Hashtbl.add t.names name t.tags;
      t.tags
    
let set t resname res = 
  printf "Set %s : %s" (List.fold_left (fun s a -> s^"."^a) "" resname ) res;
  print_newline (); 
  let resname = List.rev resname in 
  match resname with
    name :: tail ->
      Printf.printf "SEt name %s\n" name;
      let name = add_name t name in
      let b = 
        try
          Hashtbl.find t.table name
        with
          Not_found ->
            let b = Branch (ref []) in
            Hashtbl.add t.table name b;
            b
      in
      let rec iter list b =
        match list, b with
        | [] , Branch bb -> 
            bb := (1,Leaf res) :: !bb
        | name :: tail, Branch bb -> 
            let name = add_name t name in
            begin
              match !bb with
              | (n, b) :: _ when n = name ->
                  iter tail b
              | _ -> 
                  let b = Branch (ref []) in
                  bb := (name,b) :: !bb;
                  iter tail b
            end
        | _ -> assert false
      in
      iter tail b
  | _ -> failwith "Invalid Xrm entry"

exception Found of string
let rec treeget b query i star =
  Printf.printf "treeget i=%d\n" i;
  match b with
  | Leaf _ -> assert false
  | Branch bb ->
      if i >= 0 then
        let rec iter list =
          match list with
            (name,b) :: tail ->
              begin
                if name = 0 then 
                  begin
                    print_string "Star\n";
                    try
                      treeget b query i true
                    with
                      Not_found -> ()
                  end
                else
                if name = 1 && star then (* ending star *)
                  match b with
                    Leaf res -> raise (Found res)
                  | _ -> assert false
                else
                if name = query.(i) then
                  try
                    Printf.printf  "good name = query %d\n" name;
                    treeget b query (i-1) false
                  with
                    Not_found -> ()
                else
                if star then
                  for i = i-1  downto 0 do
                    if name = query.(i) then
                      try
                        treeget b query (i-1) false
                      with
                        Not_found -> ()
                  done
                else
                  Printf.printf "bad name = %d\n" name
              end;
              iter tail
          | [] -> raise Not_found
        in
        iter !bb
      else
      (* We must find a one or an ending star in the list *)
      let rec iter list =
        match list with
          (_,Leaf res) :: _ -> raise (Found res)
        | _ :: tail -> iter tail
        | [] -> raise Not_found
      in
      iter !bb
      
  
let get t resname = 
  let resname = List.rev resname in
  match resname with
    name :: tail ->
      let num = Hashtbl.find t.names name in
      let tree = Hashtbl.find t.table num in
      let query = 
        Array.of_list 
          (List.rev 
            (List.map (fun name ->
                try
                  Hashtbl.find t.names name
                with
                  Not_found -> 0) tail))
      in
      begin
        try
          treeget tree query (Array.length query - 1) false;
          raise Not_found
        with
          Found s -> s
      end
  | _ -> failwith "Invalid Xrm query"
      

let load t filename =
  let set t all res =
    try
      set t all res
    with
      Failure _ -> ()
  in
  let ic = open_in filename in
  let lexbuf = Lexing.from_channel ic in
  let rec others list =
    let token = next lexbuf in
    match token with
      STAR -> others ("*"::list)
    | NAME name -> others (name :: list)
    | POINT -> others list
    | SEMI -> List.rev list, ressource lexbuf
    | _ -> [], ""
  in
  let rec iter () =
    let token = first lexbuf in
    match token with
    | STAR -> 
        let all,res = others ["*"] in
        set t all res;
        iter ()
    | POINT -> iter ()
    | NAME name -> 
        let all,res = others [name] in
        set t all res;
        iter ()
    | _ -> ()
  in
  iter ()
  
  (* Not implemented yet *)
let save t filename = 
  Printf.printf "--";
  Hashtbl.iter (fun name num ->
      Printf.printf "%s ---> %d\n" name num) t.names;
  Hashtbl.iter (fun name b ->
      let rec iter list b =
        match b with
          Branch bb ->
            List.iter (fun (name,b) ->
                iter (name :: list) b) !bb
        | Leaf res ->
            List.iter (fun name ->
                print_char '.';
                print_int name) list;
            print_string res;
            print_newline ()
      in
      iter [name] b
  ) t.table;
  Printf.printf "--";
  print_newline () 
  
  }