(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)


open Jtypes;;

let ack_stubs stubs thread_stub =
  List.iter (fun stub -> 
      let locator = match stub.locator with None -> assert false
            | Some l -> l in
    stub.position <- Space.local_space;
    stub.gc_stub <- ref LocalStub;
    locator.lctr_age <- locator.lctr_age + 1;
    locator.lctr_weakspace <- Space.local_id;
    locator.lctr_weaknum <- 
       Whashtbl.get_direct Space.stubs_table locator.lctr_uid;
    ) stubs;
  Channel.reply_to thread_stub stubs

let unmarshal space ts modules m dest =
  let rec unmarshal_loc dest =
    fun (LocationBloc bloc) ->
      Jprims.unmarshal_location dest bloc.location bloc.loc_module
        (List.map (function (module_data,interface) ->
          Jprims.set_interface module_data interface;
          module_data) bloc.modules) 
        bloc.instance_vars;
      Array.iter Jprims.unmarshal_def bloc.automatas;
      Array.iter Jprims.unmarshal_thread bloc.threads;
      Array.iter (unmarshal_loc bloc.location) bloc.locations;
  in
  unmarshal_loc dest m

let marshal dest_space location =
  let modules = ref [] in
  let stubs = ref [] in
  let rec marshal_loc location =
    stubs := location :: !stubs;
    let (thread_stubs, defs, locs, module_data, attachements, vars) =
      Jprims.marshal_location location
    in
    Array.iter (fun stub -> stubs := stub::!stubs) defs;
    Array.iter (fun stub -> stubs := stub::!stubs) thread_stubs;
    let threads = Array.map (fun stub -> 
      stub, Jprims.marshal_thread stub) thread_stubs
    in
    let defs = Array.map (fun stub -> stub, Jprims.marshal_def stub) defs in
    let _ = Debug.print1 's' "Migrate %d attachment modules" 
        (List.length attachements)
    in
    let attachements = 
      List.map (function module_data -> 
        Debug.print1 's' "Attached %s\n" module_data.Reloc.module_name;
        if not (List.mem module_data.Reloc.module_name !modules) then
          modules := module_data.Reloc.module_name :: !modules;
        (module_data,Jprims.get_interface module_data)
          ) attachements in
    let locs = Array.map (fun stub -> marshal_loc stub) locs in
    location.local <- RemoteObject;
    if not (List.mem module_data.Reloc.module_name !modules) then
      modules := module_data.Reloc.module_name :: !modules;
    LocationBloc
      { location = location;
        threads = threads;
        automatas = defs;
        locations = locs;
        modules = attachements;
        loc_module = module_data;
        instance_vars = vars }
  in
  let m = marshal_loc location in
  !modules, m, !stubs

let migrate orig dest =
  let dest_space = dest.position in
  if dest_space.space_state = Down then Halt.halt_location orig else
  match dest.local with
    FailedObject _ -> Halt.halt_location orig
  | RemoteObject -> 
      let (modules,m,stubs) = marshal dest_space orig in
      let waiter_thread =
        Thread.create (fun () ->
          Jprims.wait_on_stub ();
          Jprims.dummy stubs; (* pour maintenir les pointeurs locaux sur 
                                 les stubs *)
          Thread.exit ()
            ) () 
      in
        Space.send dest_space 
          (Jmarshal.marshal dest_space
             (LocationMigration (modules, m, dest, stubs,
                                 Jprims.get_thread_stub waiter_thread)));
      List.iter (fun stub -> stub.position <- dest_space) stubs
          
  | _ -> assert false
      
let go dest = Jprims.go dest

let _ =
  let go_thread =
    Thread.create
      (fun () ->
        while true do
          let action = Jprims.sleep () in
          Printexc.catch (fun () -> 
            match action with
              Migrate_location(orig,dest) -> migrate orig dest
            | Kill_location location -> Halt.halt_location location
            | Kill_space(space,state) -> Halt.halt_space space state
                  ) ();
          Jprims.set_critical_section false;
        done)
      ()
  in
  Jprims.set_go_thread go_thread

let threads = ref 1

