(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)


open Jtypes;;
open Space;;
open Dgc


let find_stub_exc dest_space stub_id =
  match stub_id with
    DirectId num ->
      begin  
        match dest_space.scions_table.(num) with
          None -> raise Not_found
        | Some gcscion ->
            
            if gcscion.gcscion_generator then
              begin
                gcscion.gcscion_bck_mark <- None;
                gcscion.gcscion_generator <- false;
              end;
            gcscion.gcscion_next_stamp <- rootstamp;
            gcscion.gcscion_fwd_stamp <- rootstamp;
            let stub = gcscion.gcscion_target in
            (match stub.local with
              RemoteObject -> (* Object has moved --> short-cut *)
                Jmarshal.send dest_space (UpdateStub stub);
            | _ -> ()
            );
            stub
      end
  | IndirectId (num,uid) ->
      let stub =
        try
          match Whashtbl.find_direct stubs_table num with
            None -> raise Not_found
          | Some stub ->
              let stub_locator = match stub.locator with
                  None -> assert false | Some s -> s in
              if stub_locator.lctr_uid = uid then 
                begin
                  Debug.print1 's' "Direct Locator %d\n" num;
                  stub
                end
              else
                raise Not_found
        with
          Not_found ->
            match
              Whashtbl.find stubs_table uid 
            with
              None -> 
                Whashtbl.remove stubs_table uid;
                raise Not_found
            | Some stub -> 
                begin
                  match stub.local with
                    FailedObject _ ->
            (* send an exception message to update the locator *)
                      Debug.print0 's' "Update stub\n";
                  | _ -> ()
                end;
                stub
      in
      Jmarshal.send dest_space (UpdateStub stub);
      stub

let handler (space : space) inc =
  try
    Debug.print "New handler";
    while true do
      let (message:message) = input_value inc in
      Debug.print0 'h' "new message";
      match message with
        Dgc_msg (ts,msg) -> 
          Debug.print0 'h' "DGC message";
          if ts < space.last_ts then raise Junmarshal.MessageRefused;
          space.last_ts <- ts;
          Dgc.handler space msg
      | Join_msg (ts,buffer,locators) -> 
          Jprims.critical_section  (
          function () ->
            try
              let message = 
                Junmarshal.unmarshal space ts buffer locators in
              Debug.print1 'h' "message unmarshaled : %d\n" ts;
              match message with
                GetService (name,vt, thread_stub) ->
                  Debug.print0 'h' "GETSERVICE";
                  Channel.reply_to thread_stub
                    (try Reply (Services.getLocalService name (Obj.magic vt)) with
                      e -> Exception e)
              | Synchrone (stub_id, ord, arg, thread_stub) ->
                  Debug.print0 'h' "SYNCHRONE";
                  Channel.hard_send_asynchrone (find_stub_exc space stub_id)
                    ord (Obj.repr (thread_stub,arg))
              | Invoke (stub_id,methode,label,args,thread_stub) ->
                  Debug.print0 'h' "INVOKE";
                  Channel.invoke_no_reply (find_stub_exc space stub_id)
                    methode label args thread_stub
              | Asynchrone (stub_id, ord, arg) ->
                      Debug.print0 'h' "ASYNCHRONE";
                      let stub = find_stub_exc space stub_id in
                      Jprims.test_channel "direct" stub;
                  Channel.hard_send_asynchrone stub ord arg
              | UpdateStub stub -> ()
              | LocationMigration (modules, m, dest, stubs, thread_stub) ->
                  Debug.print0 'h' "MIGRATION";
                  Go.ack_stubs stubs thread_stub;
                  begin
                    match dest.local with
                      FailedObject _ -> 
                        List.iter (fun stub -> Halt.halt_stub stub) stubs

                    | LocalLocation local ->
                        Debug.print0 'h' "migration to local";
                        Go.unmarshal space ts modules m dest

                    | RemoteObject ->
                        begin
                          let dest_space = dest.position in
                          let waiter_thread =
                            Thread.create (fun () ->
                              Jprims.wait_on_stub ();
                              Jprims.dummy stubs;
                              Thread.exit ()
                                ) () 
                          in
                          Space.send dest_space 
                            (Jmarshal.marshal dest_space
                               (LocationMigration 
                                  (modules, m, dest, stubs,
                                   Jprims.get_thread_stub waiter_thread)));
                          List.iter (fun stub -> stub.position <- dest_space) 
                            stubs
                        end
                    | _ -> assert false
                  end
              | ReplyThread (stub_id, v) -> 
                  Debug.print0 'h' "REPLY";
                  Channel.reply_to (find_stub_exc space stub_id) v; ()
            with
              e -> Debug.print1 'h' "Exception <%s>\n" (Printexc.to_string e);
                )
    done
  with
    e -> Debug.print1 'h' "Exception <%s>\n" (Printexc.to_string e);
      Debug.print0 'h' "End_of_file ...\n";
      Jprims.fail_space space
