(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)


open Jtypes;;

external dump_heap : unit -> unit = "dump_heap"

(* extern.c *)

external cache_create: int -> cache = "cache_create"
external cache_clean: cache -> unit = "cache_clean"

external marshal_message :
  'a -> cache -> string * Obj.t array = "join_marshal_message";;

(* intern.c *)

external unmarshal_message :
  string -> Obj.t array -> cache -> 'a = "join_unmarshal_message";;

(* scheduler.c *)
external sleep': unit -> 'a = "thread_sleep"
external wait_on_stub' : unit -> 'a = "stub_thread_wait";;
external wait_on_gc': unit -> unit = "thread_wait_gc";;

external exit_wait_on_stub' : unit -> 'a = "exit_wait_on_stub";;
external gen_thread_in :
    location_stub -> ('a -> unit) -> Thread.thread = "gen_thread_in";;
      
external reply_thread_stub : thread_stub -> 'b -> unit = "stub_thread_reply";;
external set_go_thread : Thread.thread -> unit = "set_go_thread";;
external schedule_go_thread' : unit -> unit = "schedule_go_thread";;
external extract_location_threads :
  location_stub -> thread_stub array = "extract_location_threads";;
external marshal_thread :
  thread_stub -> thread_marshalled = "marshal_thread";;
external unmarshal_thread :
  thread_stub * thread_marshalled -> unit = "unmarshal_thread";;
external self_location : unit -> location_stub = "thread_self_location";;
external set_location : location_stub -> unit = "thread_set_location";;
external print_stats: unit -> unit = "print_stats"
external trigger_def_and_wait' : automata_stub * int * 'a -> 'b = "trigger_def_and_wait";;
external exit_trigger_def_and_wait' : automata_stub * int * 'a -> 'b = "exit_trigger_def_and_wait";;
external server : unit -> unit = "set_server"
external kill_all_threads : unit -> unit = "kill_all_threads"

       (* objects *)
external copy: Obj.t -> int -> int -> Obj.t = "obj_copy"
external create_object_stub : space ->  Mop.bucket array -> int -> 
  Obj.t = "create_object_stub"
external set_replicat: Obj.t -> Obj.t -> unit = "set_replicat"
external create_new_object : space -> location_stub -> int -> Reloc.module_data ->
  Obj.t = "create_new_object";;
external set_location_module : < .. > -> Reloc.module_data -> unit =
  "set_location_module"
external get_location_module: < .. > -> Reloc.module_data =
  "get_location_module"
       (* Threads *)
external get_thread_stub : Thread.thread -> thread_stub = "get_thread_stub";;
external self_stub : unit -> thread_stub = "self_stub";;
external reuse_thread : ('a -> unit) -> 'a -> thread_stub ->
  unit = "reuse_thread"
external extract_location_threads : location_stub -> thread_stub array 
    = "extract_location_threads"
       (* Locations *)
external create_new_location : space -> location_stub -> Reloc.module_data ->
  location_stub = "create_new_location";;
external create_first_location : space -> unit -> unit -> 
  location_stub = "create_new_location";;
external marshal_location : location_stub -> thread_stub array * 
    automata_stub array * location_stub array * Reloc.module_data
    * (Reloc.module_data list) * (Obj.t array)
    = "marshal_location";;
external unmarshal_location : location_stub -> location_stub -> Reloc.module_data ->
  Reloc.module_data list -> (Obj.t array) ->
  unit = "unmarshal_location";;
external go' : location_stub -> unit = "go_location";;
external kill_location' : location_stub -> 'a = "kill_location";;
external migrate' : location_stub * location_stub -> unit = "migrate";;

       (* interfaces *)
external get_interface: Reloc.module_data -> interface = "get_interface"
external set_interface: Reloc.module_data -> interface -> unit
    = "set_interface"

        (* channels *)
external test_channel': 'a -> string -> unit = "test_good_channel"
external create_empty_automata : location_stub -> 'a -> int -> int array -> bool array -> 'b
    = "create_empty_automata";;
external fill_def : automata_stub -> process array -> 'a ->
  unit = "fill_def";;
external trigger_def' : automata_stub * int * 'a -> unit = "trigger_def";;
external unmarshal_def : automata_stub * automata_marshalled -> 
  unit = "unmarshal_def";;
external marshal_def : automata_stub -> automata_marshalled = "marshal_def";;


external breakpoint : int -> 'a -> unit = "breakpoint"

(* Wrappers : C functions which call schedule_thread() must be called from
   a ML wrapper with exactly one parameter *)
let wait_on_stub () = wait_on_stub' ()
let trigger_def_and_wait args = trigger_def_and_wait' args
let exit_wait_on_stub () = exit_wait_on_stub' ();;
let exit_trigger_def_and_wait args = exit_trigger_def_and_wait' args
let trigger_def args = trigger_def' args
let schedule_go_thread () = schedule_go_thread' ();;

external set_critical_section:  bool -> bool = "set_critical_section"


external fail_space' : space -> 'a = "fail_space"
external halt_space' : space -> 'a = "halt_space"
let fail_space space =
  set_critical_section true;
  fail_space' space

let halt_space space =
  set_critical_section true;
  halt_space' space

let go l =  (* on entre immediatement dans une section critique. En fait
  le thread de migration va etre schedule dans location_go et la section 
               critique sera quittee dans cet autre thread *)
  set_critical_section true;
  go' l


let get_stub_locator (stub: generic_stub) =
  match stub.locator with
    None -> failwith "No locator"
  | Some locator -> locator

let critical_section f =
  let old = set_critical_section true in
  try
    let r = f () in
    set_critical_section  old;
    r
  with
    e -> 
      set_critical_section old;
      raise e

let wait_on_gc ()= wait_on_gc' ()
let test_channel c s = test_channel' s c

let kill_location location = kill_location' location

external register_send_sync : Obj.t -> Obj.t -> unit = "register_send_sync"
let migrate pair = migrate' pair
let sleep () = sleep' ()

let thread_create_in l fn arg =
  gen_thread_in l
    (fun () ->
      try
        Printexc.print fn arg; Thread.exit ()
      with x ->
        flush stdout; flush stderr; Thread.exit())
;;

external dummy : 'a -> unit = "%identity"
external wait_on_fail' : 'a -> unit = "suspend_on_fail"
let wait_on_fail stub = wait_on_fail' stub

external extract_stub : 'a -> generic_stub = "extract_stub"

external remove_from_father : generic_stub -> unit = "remove_from_father"
  
external print_obj: 'a -> unit = "print_obj"