(***********************************************************************)
(*                                                                     *)
(*                    JOCAML                                           *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(*         File written in the Objective-CAML language                 *)
(***********************************************************************)


(*
  Ajouter au compilateur l'optimisation concernant les types concrets de la 
forme:

type t =
  ... constructeurs simples ...
| C of t' ou t' est un type pointeur (record en particulier)

Ajouter le dernier ; dans les records construits.

*)

type port = int;;
type time = float;;
type process = Obj.t -> unit;; 
type timestamp = int;;
type interface
type cache
type code_t
type env_t

type processes
type structure =
    { nmult : int;
      nchans : int;
      tests : int; (* all tests: one structure for each channel *)
    } 

type thread_marshalled;;
type automata_marshalled;;
type space_id = Unix.inet_addr * port * time
type uid = space_id * int

type in_connection = 
    NoHandler
  | Handler of  { channel : in_channel;
                  thread : Thread.thread
                }

type out_connection =
    NoConnection
  | Connecting of Thread.thread
  | Connected of { channel : out_channel;
                   thread : Thread.thread
                 } 

type meth = string * Mop.label * int (* name, nparams, label ?? *)

type object_type =
    Location
  | Object of space_id * Mop.class_id * meth list * int
  | Automata
  | Thread

type locator =
    {
    mutable lctr_age: int;
    mutable lctr_weakspace: space_id;
    mutable lctr_weaknum: int;
    lctr_uid: uid;
    lctr_type: object_type;
  } 

type obj_locator =
  RelocModule of Reloc.module_data 
| DummyModule of string * int
| RelocException of string * string * int
| StubLocator of int * locator (* strong + locator *)
| SpecialMarshal of string * Obj.t
| Class_info of Mop.class_id
  
type space_state =
    Normal       (* either connected or never connected *)
  | MaybeDown    (* link broken or connection failed: may still recover *)
  | Down         (* space is down and will never recover *)

type mark_id = space_id * space_id * uid
type color = WhiteMark | GrayMark | BlackMark

type mark = {
    mark_range : int;
    mark_id : mark_id; 
    mark_gen : int;
  } 
type stamp = {
    mutable stamp_mark : mark; (* 0 *)
    stamp_color : color;
    mutable stamp_distance : int;
  } 


type father_location
type thread_stub = generic_stub
and location_stub = generic_stub
and automata_stub = generic_stub
and object_stub = generic_stub
and gc_stub = gclinks ref
and gclinks = 
    LocalStub
  | DirectStub of gcstub
  | IndirectStub of gcstub

(* we may use several gcstub, one for each chain of indirect pointers to the
   object, to be safe in case of failure. However, this would lead to 
   numerous cycles. When the cycles detector is implemented, this will be 
   tested, and kept if such cycles are collected efficiently.
*)
and gcstub_option_pointer =
    NoBackTrace
  | BackStub of {
(* <CGC> *)
    mutable gcstub_min : stamp;
    mutable gcstub_max : stamp;
    mutable gcstub_bck_mark : mark option;
    mutable gcstub_bck_trace : gcscion list; 
(* </CGC> *)
    gcstub_obj: generic_stub;
    gcstub_space : space;
    gcstub_age : int;
    gcstub_num : int;
    mutable gcstub_deleted : bool;
  } 

and gcstub = {
(* <CGC> *)
    mutable gcstub_min : stamp;
    mutable gcstub_max : stamp;
    mutable gcstub_bck_mark : mark option;
    mutable gcstub_bck_trace : gcscion_option_pointer; 
(* </CGC> *)
    gcstub_obj: generic_stub;
    gcstub_space : space;
    gcstub_age : int;
    gcstub_num : int;
    mutable gcstub_deleted : bool;
  } 

and space =
  { 
    space_id : space_id;
    mutable space_state : space_state;
  
  (* for communication *)
    mutable link_in : in_connection;
    mutable link_out : out_connection;
    queue: message Fifo.t;
    
  (* timestamps *)
  mutable next_ts: int;
    mutable last_ts: int;
    mutable current_ts : int;
    mutable threshold : int;

  (* for migration and objects *)
  ext_buckets : (Mop.class_id , Mop.bucket array) Hashtbl.t;
  mutable modules: string list;
  
  (* for Dgc *)
  scions_hash: (uid,int) Hashtbl.t;
  mutable scions_table : gcscion option array;
  mutable free_list: int array;
  mutable free_slots: int;
  mutable stubs_table : gcstub Weak.t;
  
    (* for simpler marshalling *)
  cache_in : cache;
  cache_out : cache;
  
    (* for CADMIUM *)
  mutable cdCache: cdCache option;
  }

and cdCache =    (* for CADMIUM *)
    Cache of space (* the cache space for the current space *)
  | Contents of (uid, (Obj.t * Obj.t)) Hashtbl.t (* contents of the cache itself *)
 (* object identifier, reference towards the original object, reference towards the copy *)

and gcscion_option_pointer =
    NoOtherScion
  | GcScion of {
(* <CGC> *)
      mutable gcscion_stamp : stamp;
      mutable gcscion_next_stamp : stamp;
      mutable gcscion_bck_mark : mark option;
      mutable gcscion_bck_stub : gcstub_option_pointer; (* None | gcstub *) 
      mutable gcscion_generator : bool;
      mutable gcscion_deleted : bool;
      mutable gcscion_fwd_stamp : stamp;
      mutable gcscion_next : gcscion_option_pointer; (* None | gcscion *)
      mutable gcscion_target : generic_stub; (* root *)
      mutable gcscion_stub : gc_stub; (* root *)
(* </CGC> *)
      mutable gcscion_ts : int;
      gcscion_space : space;
      mutable gcscion_num : int;
      mutable gcscion_age : int;
      mutable gcscion_marked : bool;
    } 

and gcscion =
    {
(* <CGC> *)
    mutable gcscion_stamp : stamp;
    mutable gcscion_next_stamp : stamp;
    mutable gcscion_bck_mark : mark option;
    mutable gcscion_bck_stub : gcstub_option_pointer; (* None | gcstub *) 
    mutable gcscion_generator : bool;
    mutable gcscion_deleted : bool;
    mutable gcscion_fwd_stamp : stamp;
    mutable gcscion_next : gcscion_option_pointer; (* None | gcscion *)
    mutable gcscion_target : generic_stub; (* root *)
    mutable gcscion_stub : gc_stub; (* root *)
(* </CGC> *)
    mutable gcscion_ts : int;
    gcscion_space : space;
    mutable gcscion_num : int;
    mutable gcscion_age : int;
    mutable gcscion_marked : bool;
  } 

and location_marshalled =
    LocationBloc of
      { location : location_stub;
        threads : (thread_stub * thread_marshalled) array;
        automatas : (automata_stub * automata_marshalled) array;
        locations : location_marshalled array;
        modules : (Reloc.module_data * interface) list;
        loc_module : Reloc.module_data;
        instance_vars : Obj.t array
      } 

and reply_type = 
    Reply of Obj.t 
  | Exception of exn

and stub_id =
    DirectId of int
  | IndirectId of int * uid

and protocole =
    GetService of string * Obj.t * thread_stub
  | Synchrone of stub_id * int * Obj.t * thread_stub
  | Invoke of stub_id * string * Mop.label * (Obj.t list) * thread_stub
  | Asynchrone of stub_id * int * Obj.t
  | LocationMigration of string list * location_marshalled * 
        location_stub * generic_stub list * thread_stub
  | UpdateStub of generic_stub
  | ReplyThread of stub_id * Obj.t

(* Le Dgc protocol a une particularite:
On ne doit y transmettre aucun locator. Uniquement des index dans des tables
gcscion <--> gcstub. De la sorte, on evite facilement le cout qu'est la
recherche de chaque identificateur et le marshalling. *)


and dgc_protocole =
    Life of timestamp * bool * ((int * stamp) list) * 
        ((int * bool * mark option) list)

and message =
    Dgc_msg of int * dgc_protocole
  | Join_msg of int * string * (obj_locator array)


and generic_stub = {
    mutable locator: locator option;
    mutable position : space; (* not the location *)
    mutable local: local_object;
    mutable cache: int;
    mutable gc_stub: gc_stub;
  }

and local_object = 
    RemoteObject 
  | LocalLocation of 
      {
      mutable father: father_location ;
      mutable def_weakset: generic_stub Weak.set;
      mutable loc_weakset: generic_stub Weak.set;
      code: Reloc.module_data;
      mutable modules: Reloc.module_data list;
      mutable free_stubs: generic_stub array;
      mutable free_stub_size: int;
      mutable next_free_stub: int
    }
  | LocalThread of 
      { ident: int; 
        mutable next: Thread.thread;
        mutable prev: Thread.thread
      } 
  | LocalAutomata of 
      { mutable state : int;
        structure : structure;
        processes : processes;
        location : generic_stub;
        queues : int (* in fact, one field for each queue *)
      } 
  | FailedObject of object_type

type channel = code_t * env_t * automata_stub * int;;
type action =
    Migrate_location of location_stub * location_stub
  | Kill_location of location_stub 
  | Kill_space of space * space_state
