(***********************************************************************)
(*                                                                     *)
(*                           JoCaml                                    *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 2000 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*
This is a rudimentary name server. It only gives on value for one key
request. There is no more liveness problem, since it doesn't support
storing remote references. 

Thus, JoCaml servers will only store names' hosts in the name server,
while the clients will first request for the object hosts, and then
query the host for the exact object.

To avoid name clashes, clients are supposed to add a small header to
object names, such as the user login, or a something like that ...

PROBLEMS:
- The current name server doesn't check that another name server is
running on the same multicast network.
- The current version depends on the marshaller version, we should instead
use a marshaller which is not version dependant.
*)

open Unix
open ThreadUnix
open Ns_types

let ignore v = ()
  
let _ = 
  Arg.parse
    [("-port", Arg.Int (fun i -> port := i), 
      (Printf.sprintf " <port>: set port name (default %d)" !port))]
    (fun s -> raise (Invalid_argument s)) "Simple name server" 

let s =
  let s = socket PF_INET SOCK_STREAM 0 in
  setsockopt s SO_REUSEADDR true;
  let localaddr = 
    try
      first_addr (gethostbyname localname) 
    with
    _ -> inet_addr_of_string "127.0.0.1"
  in
  (try bind s (ADDR_INET (localaddr, !port)) with _ ->
        let localaddr = inet_addr_of_string "127.0.0.1" in
        bind s (ADDR_INET (localaddr, !port)));
  listen s 10;
  s

let table = Hashtbl.create 50
  
let handler s =
  try
    let inc = in_channel_of_descr s in
    let outc = out_channel_of_descr s in
    let buf = String.create connection_header_len in
    really_input inc buf 0 connection_header_len;
    if buf <> connection_header then failwith "Bad connection header";
    let request = input_value inc in 
    begin
      match request with
        Register (key, value) ->
          Hashtbl.add table key value;
          output_value outc ()
      | Lookup key ->
          output_value outc (try Some (Hashtbl.find table key) 
            with _ -> None)
    end;
    flush outc;
    shutdown s SHUTDOWN_ALL
  with e ->
      Printf.printf  "Exception in handler: %s" (Printexc.to_string e);
      print_newline ()

let multicast_server () =
  Printf.printf "Starting multicast server ...";
  print_newline () ;
  try
    let sock = Multicast.server !mgroup !mport in
    Multicast.send sock !mgroup !mport mquery;
    let buffer = String.create 100 in
    try
      while true do
        let rs, _, _ = Thread.select [sock] [] [] 5.0 in
        match rs with
          [] -> raise Exit
        | _ -> 
            Multicast.recv sock buffer; 
            if String.sub buffer 0 mquery_len <> mquery then
              failwith (Printf.sprintf "Another server is running on %s"
                  (try
                    let (h,p) = Marshal.from_string buffer 0 in h
                  with _ -> "UNKNOWN"))
      done;
    with Exit ->
        Printf.printf "Server started"; print_newline () ;
        while true do
          Multicast.recv sock buffer;
          if String.sub buffer 0 mquery_len = mquery then
            let msg = Marshal.to_string (localname, !port) [] in
            Multicast.send sock !mgroup !mport msg;
        done
  with e ->
      Printf.printf "Name Server started without multicast (%s)"
        (Printexc.to_string e);
      print_newline ()

let _ = 
  if !use_multicast then
    ignore (Thread.create multicast_server ()); 
  Printf.printf "Name server started at port %d" !port; print_newline () ;
 while true do
    let s,_ = accept s in
    Thread.create handler s
  done
  
