(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;
open Jprims;;   
open Unix;;
open ThreadUnix;;

                   (* GLOBAL DATA *)
let (spaces : (space_id, space) Hashtbl.t) = Hashtbl.create 11
let (stubs_table : (uid, generic_stub) Whashtbl.t) = Whashtbl.create 101
let (name_server : (string, Obj.t*Obj.t) Hashtbl.t) = Hashtbl.create 11

(*let (weak_array : gc_stub Weak.array ref) = ref (Weak.array_create 4100)*)

                    (* FUNCTIONS *)

let creation_time = gettimeofday ();;
let localname = gethostname ();;
let first_addr entry = entry.h_addr_list.(0);;

let localname = (* try to bind a socket to a INET port, if not possible,
use loopback (127.0.0.1) instead... *)
  let s = socket PF_INET SOCK_STREAM 0 in
  setsockopt s SO_REUSEADDR true;  
  let localaddr = 
    try
      first_addr (gethostbyname localname) 
    with
    | _ -> inet_addr_of_string "127.0.0.1"
    in
  let localname = 
    try
      bind s (ADDR_INET (localaddr, 0));
      localname
    with
      _ -> 
        Printf.printf "WARNING: Using loopback address (127.0.0.1). you will not be able to communicate with other machines. Verify your network connection if this is not normal.";
        print_newline ();
        let localaddr = inet_addr_of_string "127.0.0.1" in
        bind s (ADDR_INET (localaddr, 0));
        "localhost"
  in
  close s;
  localname

let localaddr = 
  try
    first_addr (gethostbyname localname)
  with
  | _ -> inet_addr_of_string "127.0.0.1"
;;

let join_handler = ref (fun space inc -> ());;
let join_initialised = ref false;;

let string_of_space_id (addr,port,time) =
  (Unix.string_of_inet_addr addr)^"@"^(string_of_int port)
  ^":"^(string_of_float time)

let string_of_uid = function
    (space_id,ts) ->
      "UID ("^(string_of_int ts)^" "^(string_of_space_id space_id)^")"

let create_port port =
  handle_unix_error
    (fun () ->
      let s = socket PF_INET SOCK_STREAM 0 in
      let _ =
        setsockopt s SO_REUSEADDR true;
        bind s (ADDR_INET (localaddr, port));
        listen s 5;
        ()
      in
      let saddr = getsockname s in
      match saddr with
        ADDR_INET (_, port) -> port, s
      | _ -> failwith "not ADDR_INET socket") ()

let first_link = create_port 0;;
let local_id = localaddr, fst first_link,creation_time;;

let local_space = 
  { space_id = local_id;
    space_state = Normal;
    
    queue = Fifo.create ();
    link_in = NoHandler;
    link_out = NoConnection;

    next_ts = 1;
    last_ts = 0;
    current_ts = 0;
    threshold = 0;

    ext_buckets = Hashtbl.create 13;
    modules = [];

    scions_hash = Hashtbl.create 13;
    scions_table = [||];
    free_list = [||];
    free_slots = 0;
    stubs_table = Weak.create 0;

    cache_in = cache_create 1;
    cache_out = cache_create 1;
    cdCache = None; (* for CADMIUM *)
  }
let local_location = Jprims.create_first_location local_space () ();;

let _ =
  Hashtbl.add spaces local_id local_space
      
let get_space (space_id:space_id) =
  Debug.print1 's' "get space %s\n" (string_of_space_id space_id);
  try Hashtbl.find spaces space_id with
    Not_found ->
      let space =
        { space_id = space_id;
          space_state = Normal;
          queue = Fifo.create ();
          last_ts = 0;
          ext_buckets = Hashtbl.create 13;
          next_ts = 1;
          current_ts = 0;
          modules = [];
          threshold = 0;
          scions_hash = Hashtbl.create 13;
          scions_table = [||];
          free_list = [||];
          free_slots = 0;
          link_in = NoHandler;
          link_out = NoConnection;
          stubs_table = Weak.create 100;
          cache_in = cache_create 512;
          cache_out = cache_create 512;
          cdCache = None; (* for CADMIUM *)
        } in
      Hashtbl.add spaces space_id space;
      space

let sender_work (space,outc)  =
  while true do
    try
      while true do
        Jprims.breakpoint 1 ();
        set_critical_section true;
        let message = Fifo.take space.queue in
        Jprims.breakpoint 2 ();
        
        set_critical_section false;
        Debug.print "send'";
        let s = Marshal.to_string  message []  in
        let l = String.length s in
        Debug.print1 's' "Space.send : %d\n" l;
        output_string outc s;
        Debug.print "Space.send : output value OK "
      done
    with
      Fifo.Empty -> 
        Jprims.breakpoint 3 ();

        flush outc;
        Thread.sleep ();
        Jprims.breakpoint 0 ();
        Debug.print "Space.send : woke up "
    | _ -> 
        Debug.print "Space send : failure of space";
        Jprims.fail_space space
  done

let links = ref [first_link];;
let threads = ref 0

let rec accept s =
  try
    ThreadUnix.accept s
  with
    Unix_error(EAGAIN,_,_) -> 
      Debug.print0 's' "accept: Try again\n";
      accept s
  | e -> 
      accept s

let old_open_connection = open_connection
  
let rec open_connection addr = 
  try
    old_open_connection addr
  with e -> Printf.printf "Exc %s" (Printexc.to_string e); 
      print_newline ();
      raise e
      
      
let rec wait_on_socket s =
  incr threads;
  Thread.create
    (fun () ->
       while true do
        try
          let (s, _) = accept s in
           let inc = in_channel_of_descr s in
           let outc = out_channel_of_descr s in
           let space_id = input_value inc in
           output_value outc local_id;
           flush outc;
           if input_value inc then
             begin
               let modules = input_value inc in
               Jprims.set_critical_section true;
               let space = get_space space_id in
               space.modules <- modules;
               output_value outc (match space.link_out with
                 NoConnection ->
                   space.link_out <- 
                      Connected { channel = outc;
                                  thread = (Thread.create sender_work (space,outc))
                                }; 
                   true
               | _ -> false);
               flush outc;
               begin
                 match space.link_in with
                   Handler handler ->
                     space.link_in <- NoHandler;
                     Thread.kill handler.thread
                 | _ -> ()
               end;
               space.link_in <-
                  Handler
                    { channel = inc;
                      thread = Thread.create
                        (fun () ->
                          Debug.print0 's' "Starting join handler\n";
                          !join_handler space inc
                            ) ()
                    };
               space.space_state <- Normal;
               Jprims.set_critical_section false;
               Debug.print1 's' "Link open from %s\n"
                 (string_of_space_id  space.space_id)
             end;
         with
           e -> Debug.print "Exception in wait_on_socket"; ()
       done)
    ()

and
(* inside join_critical *)
    connect_to space =
  try
    let (addr,port,time) = space.space_id in
    let (inc, outc) = open_connection (ADDR_INET (addr, port))
    in
    output_value outc local_id;
    flush outc;
    let local_id = input_value inc in
    if local_id = space.space_id then (* OK *)
      begin
        output_value outc true;
        output_value outc (Reloc.module_list ());
        flush outc;
        if input_value inc then
          begin
            match space.link_in with
              Handler handler -> 
                space.link_in <- NoHandler;
                Thread.kill handler.thread
            | _ -> 
                space.link_in <-
                   Handler
                     { channel = inc;
                       thread = Thread.create
                         (fun () ->
                           Debug.print0 's' "Starting join handler\n";
                           !join_handler space inc
                             ) ()
                     }; 
          end;
        space.link_out <- 
           Connected { channel = outc;
                       thread = (Thread.create sender_work (space,outc))
                     }; 
        space.space_state <- Normal;
      end
    else
      begin
        output_value outc false;
        Jprims.halt_space space;
      end
  with
    e -> 
      Debug.print "Error in connect_to"; 
      fail_space space;
      raise e
        
and first_connect_to space_addr =
  Debug.print "Connecting to space ......";
  Printf.printf "C"; print_newline ();
  try
    let (inc, outc) = 
      open_connection (ADDR_INET (fst space_addr,snd space_addr))
    in
    output_value outc local_id;
    flush outc;
    let space_id = input_value inc in
    let space = get_space space_id in
    if space.link_out = NoConnection then
      begin
        output_value outc true;
        output_value outc (Reloc.module_list ());
        flush outc;
        if input_value inc then
          begin
            match space.link_in with
              Handler handler -> 
                space.link_in <- NoHandler;
                Thread.kill handler.thread
            | _ -> 
                space.link_in <-
                   Handler
                     { channel = inc;
                       thread = Jprims.thread_create_in local_location
                         (fun () -> 
                           Debug.print0 's' "Starting join handler\n";
                           !join_handler space inc
                             ) ()
                     }; 
          end;
        space.link_out <- 
           Connected { channel = outc;
                       thread = (Jprims.thread_create_in local_location
                                   sender_work (space,outc))
                     }; 
        space.space_state <- Normal;
      end;
    space
  with
    e -> 
      Debug.print "Error in first connect to"; 
      raise e

let hard_send space message =
  Debug.print1 's' "hard send to %s\n" (string_of_space_id space.space_id);
  match space.space_state with
    Down -> 
      Debug.print0 's' "Space is down"
  | _ ->
      Fifo.put space.queue message;
      match space.link_out with
        NoConnection ->
          space.link_out <- Connecting
               (Jprims.thread_create_in local_location 
                  (fun () ->
                    connect_to space
                      ) ());
      | Connecting thread -> ()
      | Connected connection ->
          Debug.print "Wake up";
          Thread.wakeup connection.thread;
          Debug.print "OK"
        
let send space message =
  Jprims.critical_section (function () -> 
    hard_send space message)

let uid_counter = ref 0
let new_uid () = 
  incr uid_counter;
  (local_id,!uid_counter)

let create_new_port port =
  let (port, s) = create_port port in 
  wait_on_socket s; port, s
;;

let new_port port =
  if List.mem_assoc port !links then
    failwith "Attempt to bind twice the same port"
  else links := create_new_port port :: !links
;;
    

let _ =
  Jprims.set_location local_location;
  wait_on_socket (snd first_link);;


let set_cache space = local_space.cdCache <- (Some (Cache space)) (* for CADMIUM *)


let find_stub num uid =
  Debug.print0 'a' "Direct ?";
  try
    match Whashtbl.find_direct stubs_table num with
      None -> raise Not_found
    | Some stub ->
        let stub_locator = match stub.locator with
            None -> assert false
          | Some s -> s in
        if stub_locator.lctr_uid = uid then 
          begin
            Debug.print1 's' "Direct Locator %d\n" num;
            stub
          end
        else
          raise Not_found
  with
    Not_found ->
      match
        Whashtbl.find stubs_table uid 
      with
        None -> 
          Whashtbl.remove stubs_table uid;
          raise Not_found
      | Some stub -> 
          stub

let space_of_addr addr port = first_connect_to (addr,port)

